/*=============================================================================
Vitalnix - User Management Suite
sysprog/shared.c - Shared functions for the ACCDB system utilities
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2004
  -- License restrictions apply (Dual FAL/LGPL2).
  -- For details see doc/FAL.txt and doc/LGPL2.txt.
  $Id: shared.c,v 1.7 2004/03/11 14:53:03 hirogen2 Exp $
=============================================================================*/
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <libHX.h>
#include "sysprog/shared.h"
#ifndef O_BINARY
#    define O_BINARY 0
#endif

typedef unsigned char byte;

//-----------------------------------------------------------------------------
int Sh_copy_file(const char *src, const char *dest) {
    char buf[MAXLNLEN];
    int dd, eax = 0, sd, l;

    if((sd = open(src, O_RDONLY | O_BINARY, 0400)) < 0) { return -errno; }
    if((dd = open(dest, O_WRONLY | O_BINARY | O_CREAT | O_TRUNC | O_EXCL)) < 0) {
        /* Already existing files will not be overwritten, thus errno must be
        reset to 0. */
        eax = errno;
        close(sd);
        if(eax == EEXIST) { return 0; }
        return -(errno = eax);
    }

    while((l = read(sd, buf, MAXLNLEN)) > 0 && write(dd, buf, l) > 0);
    close(sd);
    close(dd);
    return 1;
}

int Sh_copy_dir(const char *src, const char *dest, long uid, long gid) {
    void *dt = HXdir_open(src);
    char *fn;
    if(dt == NULL) { return 0; }

    while((fn = HXdir_read(dt)) != NULL) {
        struct stat sb;
        char fsrc[MAXFNLEN], fdest[MAXFNLEN];
        if(strcmp(fn, ".") == 0 || strcmp(fn, "..") == 0) { continue; }
        snprintf(fsrc, MAXFNLEN, "%s/%s", src, fn);
        snprintf(fdest, MAXFNLEN, "%s/%s", dest, fn);

        lstat(fsrc, &sb);
        sb.st_mode &= ~07000; // clear SUID/GUID/Sticky bits
        if(S_ISREG(sb.st_mode)) {
            Sh_copy_file(fsrc, fdest);
        } else if(S_ISDIR(sb.st_mode)) {
            HX_mkdir_p(fdest, sb.st_mode);
            Sh_copy_dir(fsrc, fdest, uid, gid);
        } else if(S_ISLNK(sb.st_mode)) {/*
            char pt[MAXFNLEN];
            memset(pt, '\0', MAXFNLEN);
            readlink(fsrc, pt, MAXFNLEN - 1);*/
        } else if(S_ISBLK(sb.st_mode) || S_ISCHR(sb.st_mode)) {
            mknod(fdest, sb.st_mode, sb.st_dev);
        } else if(S_ISFIFO(sb.st_mode)) {
            mkfifo(fdest, sb.st_mode);
        }

        chown(fdest, uid, gid);
        chmod(fdest, sb.st_mode);
    }

    HXdir_close(dt);
    return 1;
}

int Sh_count_args(const char **args) {
    int argk = 0;
    while(*args++ != NULL) { ++argk; }
    return argk;
}

void Sh_interactive(struct itab *qtabp) {
    char buf[MAXLNLEN];

    while(qtabp->fmt != NULL) {
        if(qtabp->type == 's') {
            char **ptr = qtabp->ptr;
            printf(qtabp->fmt, (*ptr != NULL) ? *ptr : "");
            fgets(buf, MAXLNLEN, stdin);
            HX_chomp(buf);
            if(*buf != '\0') { HX_strclone(ptr, buf); }
        } else if(qtabp->type == 'd') {
            int *ptr = qtabp->ptr;
            printf(qtabp->fmt, (ptr != NULL) ? *ptr : 0);
            fgets(buf, MAXLNLEN, stdin);
            HX_chomp(buf);
            if(*buf != '\0') { *ptr = strtol(buf, NULL, 0); }
        }

        if(!(qtabp->callback != NULL &&
         !qtabp->callback(qtabp->fmt, qtabp->type, qtabp->ptr))) {
            ++qtabp;
        }
    }

    return;
}

long Sh_string_to_iday(char *s) {
    struct tm td;
    time_t r;
    memset(&td, 0, sizeof(struct tm));

    if(strchr(s, '.') != NULL) {
        // DD.MM.YYYY
        sscanf(s, "%d.%d.%d", &td.tm_mday, &td.tm_mon, &td.tm_year);
    } else if(strchr(s, '/') != NULL) {
        // MM/DD/YYYY
        sscanf(s, "%d/%d/%d", &td.tm_mon, &td.tm_mday, &td.tm_year);
    } else if(strchr(s, '-') != NULL) {
        // YYYY-MM-DD (ISO 8601) format
        sscanf(s, "%d-%d-%d", &td.tm_year, &td.tm_mon, &td.tm_mday);
    } else {
        return -1;
    }

    r = mktime(&td);
    if(r == -1) { return -1; }
    return (unsigned long)r / 86400;
}

int Sh_only_digits(char *p) {
    if(p == NULL) { return 0; }
    while(*p != '\0') {
        if(!isdigit(*p++)) { return 0; }
    }
    return 1;
}

//==[ End of file ]============================================================
