/*=============================================================================
Vitalnix/ACCDB - Unified Account Database
sysprog/usermod.c - Modify a user account
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2004
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  $Id: usermod.c,v 1.7 2004/03/11 14:56:06 hirogen2 Exp $
=============================================================================*/
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "sysprog/shared.h"

typedef unsigned char byte;

enum {
    E_SUCCESS = 0,
    E_OTHER,       // other error, see errno
    E_OPEN,        // unable to open back-end module or DB
    E_NOEXIST,     // user does not exist
    E_UID_USED,    // UID already used and -o was not specified
    E_NAME_USED,   // new user name (-n) already exists
    E_UPDATE,      // db->usermod() did not return ok
    E_CLOSE,       // db->close() did not return ok
};

enum {
    AOP_LOCK = 1,
    AOP_UNLOCK,
};

struct {
    char *lname, *new_lname;
    long new_uid;
    char *new_igrp, *new_sgrp, *new_gecos, *new_homedir, *new_shell,
     *new_passwd;
    long new_expire, new_inactive;

    // usermod internal
    char *ac_after, *ac_before;
    int dup, inter, lock, mov;
} opt = {
    lname: NULL, new_lname: NULL, new_uid: -1, new_igrp: NULL, new_sgrp: NULL,
    new_gecos: NULL, new_homedir: NULL, new_shell: NULL, new_passwd: NULL,
    new_expire: -1, new_inactive: -1,
    ac_after: NULL, ac_before: NULL, dup: 0, inter: 0, lock: 0, mov: 0,
};

static int check_login(const char *);
static int get_options(int *, const char ***);
static int iconfig_lname(const char *, char, void *);
static void read_config(void);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    struct accdb_module *db;
    struct accdb_user mmask, smask, result;
    int eax, rv = E_SUCCESS;

    memset(&mmask, 0, sizeof(struct accdb_user));
    memset(&smask, 0, sizeof(struct accdb_user));
    memset(&result, 0, sizeof(struct accdb_user));
    mmask.uid = mmask.gid = -1;
    mmask.last_change = mmask.keep_min = mmask.keep_max = mmask.warn_age = -1;
    mmask.expire = mmask.inactive = -1;

    read_config();
    if(!(eax = get_options(&argc, (const char ***)&argv))) { return E_OTHER; }

    if((db = accdb_load(module_path)) == NULL) {
        fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
         module_path, strerror(errno));
        return E_OPEN;
    }

    if((eax = db->open(&db->state)) <= 0) {
        fprintf(stderr, "Could not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_OPEN;
        goto __main__close_accdb;
    }

    // ----------------------------------------
    smask.lname = opt.lname = *argv;
    smask.uid = smask.gid = -1;

    if((eax = db->userinfo(db->state, &smask, &result, 1)) < 0) {
        fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_OTHER;
        goto __main__close_backend;
    }

    if(eax == 0) {
        fprintf(stderr, "User \"%s\" does not exist\n", smask.lname);
        rv = E_NOEXIST;
        goto __main__close_backend;
    }

    // ----------------------------------------
    if(opt.inter) {
        struct itab t[] = {
            {"Login name\n[%s] > ", 's', &mmask.lname, iconfig_lname},
            {"User ID (-1 for next free)\n[%d] > ", 'l', &mmask.uid, NULL},
            {"Initial group (or GID)\n[%s] > ", 's', &opt.new_igrp, NULL},
            {"Supplemental groups (and/or GIDs), seperated by comma\n[%s] > ",
             's', &opt.new_sgrp, NULL},
            {"Comment field (real name, etc.)\n[%s] > ",
             's', &mmask.gecos, NULL},
            {"Home directory\n[%s] > ", 's', &mmask.home, NULL},
            {"Shell\n[%s] > ", 's', &mmask.shell, NULL},
            {NULL},
        };
        memcpy(&mmask, &result, sizeof(struct accdb_user));
        if(opt.new_igrp == NULL) {
            char buf[16];
            snprintf(buf, 16, "%ld", result.gid);
            opt.new_igrp = (result.igrp != NULL) ?
             result.igrp : HX_strdup(buf);
        }
        Sh_interactive(t);
    }

    if(opt.new_igrp != NULL) {
        if(Sh_only_digits(opt.new_igrp)) {
            mmask.gid  = strtoul(opt.new_igrp, NULL, 0);
            mmask.igrp = NULL;
        } else {
            mmask.gid  = -1;
            mmask.igrp = opt.new_igrp;
        }
    }

    if(opt.new_lname    != NULL) { mmask.lname    = opt.new_lname;    }
    if(opt.new_uid      != -1)   { mmask.uid      = opt.new_uid;      }
    if(opt.new_gecos    != NULL) { mmask.gecos    = opt.new_gecos;    }
    if(opt.new_homedir  != NULL) { mmask.home     = opt.new_homedir;  }
    if(opt.new_shell    != NULL) { mmask.shell    = opt.new_shell;    }
    if(opt.new_passwd   != NULL) { mmask.passwd   = opt.new_passwd;   }
    if(opt.new_expire   != -1)   { mmask.expire   = opt.new_expire;   }
    if(opt.new_inactive != -1)   { mmask.inactive = opt.new_inactive; }

    if(mmask.uid != -1 && mmask.uid != result.uid && !opt.dup) {
        // If UID has changed...
        struct accdb_user quser;
        memset(&quser, 0, sizeof(struct accdb_user));
        quser.uid = mmask.uid;
        quser.gid = -1;

        if(db->userinfo(db->state, &quser, NULL, 0) > 0) {
            fprintf(stderr, "A user with UID %ld already exists."
             " Use -o to override.\n", mmask.uid);
            rv = E_UID_USED;
            goto __main__close_backend;
        }
    }

    if(mmask.lname != NULL && result.lname != NULL &&
     strcmp(mmask.lname, result.lname) != 0) {
        // ... if name changed
        struct accdb_user quser;
        memset(&quser, 0, sizeof(struct accdb_user));
        quser.lname = mmask.lname;
        quser.uid   = \
        quser.gid   = -1;

        if(db->userinfo(db->state, &quser, NULL, 0) > 0) {
            fprintf(stderr, "A user with that name (\"%s\") already"
             " exists.\n", mmask.lname);
            rv = E_NAME_USED;
            goto __main__close_backend;
        }
    }

    // ----------------------------------------
    if(opt.ac_before != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, opt.ac_before, mmask.lname, mmask.uid);
        system(cmd);
    }

    if((eax = db->usermod(db->state, &smask, &mmask)) <= 0) {
        fprintf(stderr, "Error: User updating failed: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_UPDATE;
    } else if(opt.ac_after != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, opt.ac_after, mmask.lname, mmask.uid);
        system(cmd);
    }

    // ----------------------------------------
__main__close_backend:
    if((eax = db->close(db->state)) <= 0) {
        fprintf(stderr, "Warning: DB was not cleanly closed:"
         " %s (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        // if there was some other error, keep that instead
        if(rv == E_SUCCESS) { rv = E_CLOSE; }
    }

__main__close_accdb:
    accdb_unload(db);
    return rv;
}

//-----------------------------------------------------------------------------
static int check_login(const char *ln) {
    if(*ln == '\0') {
        fprintf(stderr, "You need to provide a username\n");
        return 0;
    }
    return 1;
}

static int get_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL};
    const struct poptOption options_table[] = {
        // New, Vitalnix-usermod options
        {NULL, 'A', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &opt.ac_after,
         '\0', "Program to run after user modification", NULL},
        {NULL, 'B', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &opt.ac_before,
         '\0', "Program to run before user modification", NULL},
        {NULL, 'I', POPT_ARG_NONE, &opt.inter, '\0',
         "Interactively prompt for parameters", NULL},
        {NULL, 'M', POPT_ARG_STRING, &module_path, '\0',
         "Use a different module than \"*\" (the default)", "name"},

        // Default options
        {NULL, 'G', POPT_ARG_STRING, &opt.new_sgrp, '\0',
         "The new supplementary groups for the user", "groups"},
        {NULL, 'L', POPT_ARG_VAL, &opt.lock, AOP_LOCK,
         "Lock user account", NULL},
        {NULL, 'U', POPT_ARG_VAL, &opt.lock, AOP_UNLOCK,
         "Unlock user account", NULL},
        {NULL, 'c', POPT_ARG_STRING, &opt.new_gecos, '\0',
         "New comment for the user", "comment"},
        {NULL, 'd', POPT_ARG_STRING, &opt.new_homedir, '\0',
         "New home directory (home automatically moved)", "dir"},
        {NULL, 'e', POPT_ARG_STRING, &opt.new_expire, '\0',
         "Date when the account expires", "date"},
        {NULL, 'f', POPT_ARG_INT, &opt.new_inactive, '\0',
         "Days until account becomes inactive", "days"},
        {NULL, 'g', POPT_ARG_LONG, &opt.new_igrp, '\0',
         "The new primary group for the user", "group"},
        {NULL, 'l', POPT_ARG_STRING, &opt.new_lname, '\0',
         "New login name of the user", NULL},
        {NULL, 'm', POPT_ARG_NONE, &opt.mov, '\0',
         "Move home directory if -d is given", NULL},
        {NULL, 'o', POPT_ARG_NONE, &opt.dup, '\0',
         "Allow creating a user with non-unique UID", NULL},
        {NULL, 'p', POPT_ARG_STRING, &opt.new_passwd, '\0',
         "New password to use (encrypted form)", NULL},
        {NULL, 's', POPT_ARG_STRING, &opt.new_shell, '\0',
         "New default shell for user", "file"},
        {NULL, 'u', POPT_ARG_LONG, &opt.new_uid, '\0',
         "New numerical value of the user's ID", "uid"},
        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    while((c = poptGetNextOpt(ctx)) >= 0);
    if(c < -1) {
        fprintf(stderr, "%s: %s\n", poptBadOption(ctx, 0), poptStrerror(c));
        poptFreeContext(ctx);
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = Sh_count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    if(args == NULL || *args == NULL) {
        // loginname is mandatory
        poptPrintUsage(ctx, stderr, 0);
        poptFreeContext(ctx);
        return 0;
    }

    poptFreeContext(ctx);
    return 1;
}

static int iconfig_lname(const char *key, char type, void *ptr) {
    char **lnamep = ptr;
    if(!check_login(*lnamep)) { return 0; }
//    if(opt.new_homedir == NULL) { generate_homedir(*lnamep); }
    return 1;
}

static void read_config(void) {
    static const char *paths[] = {"sysprog/../vetc", "/etc", NULL};
    static struct rconfig_opt options_table[] = {
        {"AC_BEFORE", RCONF_STRING, &opt.ac_before, NULL},
        {"AC_AFTER",  RCONF_STRING, &opt.ac_after,  NULL},
        {NULL},
    };

    accdb_rconfig_pv(paths, "usermod.conf", options_table, RCONFIG_ONE);
    return;
}

//==[ End of file ]============================================================
